/*
 * Authors (alphabetical order)
 * - Andre Bernet <bernet.andre@gmail.com>
 * - Andreas Weitl
 * - Bertrand Songis <bsongis@gmail.com>
 * - Bryan J. Rentoul (Gruvin) <gruvin@gmail.com>
 * - Cameron Weeks <th9xer@gmail.com>
 * - Erez Raviv
 * - Gabriel Birkus
 * - Jean-Pierre Parisy
 * - Karl Szmutny
 * - Michael Blandford
 * - Michal Hlavinka
 * - Pat Mackenzie
 * - Philip Moss
 * - Rob Thomson
 * - Romolo Manfredini <romolo.manfredini@gmail.com>
 * - Thomas Husterer
 *
 * opentx is based on code named
 * gruvin9x by Bryan J. Rentoul: http://code.google.com/p/gruvin9x/,
 * er9x by Erez Raviv: http://code.google.com/p/er9x/,
 * and the original (and ongoing) project by
 * Thomas Husterer, th9x: http://code.google.com/p/th9x/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "../../opentx.h"

extern Fifo<512> telemetryFifo;

struct t_SportTx
{
	uint8_t *ptr ;
	uint32_t count ;
} SportTx ;

void telemetryPortInit(uint32_t baudrate)
{
  if (baudrate == 0) {
    USART_DeInit(USART2);
    return;
  }

  USART_InitTypeDef USART_InitStructure;
  GPIO_InitTypeDef GPIO_InitStructure;

  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIO_SPORT, ENABLE);

  GPIO_PinAFConfig(GPIO_PIN_SPORT_TXRX, GPIO_PinSource_SPORT_RX, GPIO_AF_SPORT);
  GPIO_PinAFConfig(GPIO_PIN_SPORT_TXRX, GPIO_PinSource_SPORT_TX, GPIO_AF_SPORT);

  GPIO_InitStructure.GPIO_Pin = PIN_SPORT_TX | PIN_SPORT_RX;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIO_PIN_SPORT_TXRX, &GPIO_InitStructure);
  
  GPIO_InitStructure.GPIO_Pin = PIN_SPORT_ON;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIO_PIN_SPORT_ON, &GPIO_InitStructure);
  
  GPIO_ResetBits(GPIO_PIN_SPORT_ON, PIN_SPORT_ON);
  
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPORT, ENABLE);
  
  USART_InitStructure.USART_BaudRate = baudrate;
  USART_InitStructure.USART_WordLength = USART_WordLength_8b;
  USART_InitStructure.USART_StopBits = USART_StopBits_1;
  USART_InitStructure.USART_Parity = USART_Parity_No;
  USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
  USART_InitStructure.USART_Mode = USART_Mode_Tx | USART_Mode_Rx;
  
  USART_Init(SPORT, &USART_InitStructure);
  USART_Cmd(SPORT, ENABLE);
  
  USART_ITConfig(SPORT, USART_IT_RXNE, ENABLE);

  NVIC_EnableIRQ(USART2_IRQn);
  NVIC_SetPriority(USART2_IRQn, 6);
}

void telemetryPutc(const char c)
{
  GPIO_SetBits(GPIO_PIN_SPORT_ON, PIN_SPORT_ON);
  USART_SendData(SPORT, c);
  GPIO_ResetBits(GPIO_PIN_SPORT_ON, PIN_SPORT_ON);
}

#if !defined(SIMU)

#define USART_FLAG_ERRORS (USART_FLAG_ORE | USART_FLAG_NE | USART_FLAG_FE | USART_FLAG_PE)


void x9dSPortTxStart( uint8_t *buffer, uint32_t count )
{
	SportTx.ptr = buffer ;
	SportTx.count = count ;
	GPIOD->BSRRL = 0x0010 ;		// output enable
//	USART2->SR = ~USART_SR_TC ;
	USART2->CR1 |= USART_CR1_TXEIE ;
}


extern "C" void USART2_IRQHandler()
{
  uint32_t status;
  uint8_t data;

  status = SPORT->SR;

	if ( status & USART_SR_TXE )
	{
		if ( SportTx.count )
		{
			USART2->DR = *SportTx.ptr++ ;
			if ( --SportTx.count == 0 )
			{
				USART2->CR1 &= ~USART_CR1_TXEIE ;	// Stop Tx interrupt
				USART2->CR1 |= USART_CR1_TCIE ;	// Enable complete interrupt
			}
		}
	}
	
	if ( USART2->SR & USART_SR_TC )
	{
		USART2->CR1 &= ~USART_CR1_TCIE ;	// Stop Complete interrupt
		GPIOD->BSRRH = 0x0010 ;		// output disable
	}
  
	while (status & (USART_FLAG_RXNE | USART_FLAG_ERRORS)) {
    data = SPORT->DR;

    if (!(status & USART_FLAG_ERRORS))
      telemetryFifo.push(data);

    status = SPORT->SR;
  }
}

#endif
